import { Injectable } from '@nestjs/common';
import { SalesPriceFormulaDataService } from 'src/modules/transaction/sales-price-formula/data/services/sales-price-formula-data.service';
import { TaxDataService } from 'src/modules/transaction/tax/data/services/tax-data.service';
import { TransactionEntity } from '../../entities/transaction.entity';
import { calculateFormula } from 'src/modules/transaction/sales-price-formula/domain/usecases/managers/helpers/calculation-formula.helper';
import * as math from 'mathjs';
import {
  TransactionBundlingItemEntity,
  TransactionItemEntity,
} from '../../entities/transaction-item.entity';

@Injectable()
export class PriceCalculator {
  constructor(
    private formulaService: SalesPriceFormulaDataService,
    private taxService: TaxDataService,
  ) {}

  private initialValue(formulas: any[]) {
    const tax = {};
    for (const formula of formulas) {
      tax[`${formula.name}_value`] = null;
    }

    return tax;
  }

  async calculate(transaction: TransactionEntity) {
    const prices = [];
    // for (const item of transaction.items) {
    //   const price = await this.calculateItem(item);
    //   const priceValues = this.calculatePrice(price);

    //   prices.push(priceValues);
    // }

    for (let i = 0; i < transaction.items.length; i++) {
      const item = transaction.items[i];
      const price = await this.calculateItem(item);
      const priceValues = this.calculatePrice(price);
      prices.push(priceValues);

      if (item.bundling_items) {
        for (let b = 0; b < item.bundling_items.length; b++) {
          const bundling = item.bundling_items[b];
          bundling.total_price = bundling.item_rates ?? bundling.base_price;

          const bundlingPrice = await this.calculateItem(bundling);
          const bundlingValues = this.calculatePrice(bundlingPrice);

          Object.assign(bundling, bundlingValues);
          item.bundling_items[b] = bundling;
        }
      }

      Object.assign(item, priceValues);
      transaction.items[i] = item;
    }

    const { payment_total_dpp, ...otherValue } =
      this.sumArrayBasedOnObjectKey(prices);

    return { dpp_value: payment_total_dpp, other: otherValue, tax_datas: [] };
  }

  async calculateItem(
    transaction: TransactionItemEntity | TransactionBundlingItemEntity,
  ) {
    const itemShare = 20 / 100;
    const profitShare = 10 / 100;

    const tax = await this.taxService.taxKeyValue();
    const taxes = { ...tax, item_share: itemShare, profit_share: profitShare };
    const formulas = await this.formulaService.salesPriceFormula();
    const taxFormula = await this.taxService.getManyByOptions({});
    const shareFormulas = await this.formulaService.profitShareFormula();
    const values = {
      total: transaction.total_price,
      ...this.initialValue(taxFormula),
      ...taxes,
    };

    // const dpp = formulas.find((formula) => formula.value_for == 'dpp');
    const dpp = formulas;

    const dppValue = calculateFormula(
      dpp.formula_string,
      taxes,
      transaction.total_price,
    );

    values['dpp'] = dppValue;
    values['dpp_value'] = dppValue;

    let calledVariable = [];
    do {
      const valueFor = this.withNullValue(values, calledVariable);
      const formula = taxFormula.find(
        (formula) => `${formula.name}_value` == valueFor,
      );
      let result = null;

      try {
        result = math.evaluate(formula.formula_string, values);
        calledVariable = [];
      } catch (error) {
        calledVariable.push(valueFor);
        console.log(error);
      }

      values[valueFor] = result;
      // values[`${valueFor}_value`] = result;
    } while (this.containsNullValue(values));

    const itemShareFormula = shareFormulas.find(
      (f) => f.value_for == 'item_share',
    );
    values['item_share_value'] = math.evaluate(
      itemShareFormula.formula_string,
      values,
    );

    const profitShareFormula = shareFormulas.find(
      (f) => f.value_for == 'profit_share',
    );
    values['profit_share_value'] = math.evaluate(
      profitShareFormula.formula_string,
      values,
    );

    return { dpp_value: dppValue, tax_datas: values };
  }

  calculatePrice(prices: any) {
    const data = prices.tax_datas;
    const filteredObject: any = Object.keys(data)
      .filter((key) => key.endsWith('_value'))
      .reduce((acc, key) => ({ ...acc, [key]: data[key] }), {});

    const { dpp_value, item_share_value, profit_share_value, ...tax } =
      filteredObject;
    const taxes = this.sumPriceObject(tax);
    return {
      total_profit_share: item_share_value,
      total_share_tenant: profit_share_value,
      payment_total_tax: taxes,
      payment_total_dpp: dpp_value,
    };
  }

  sumArrayBasedOnObjectKey(arr) {
    return arr.reduce((acc, cur) => {
      Object.keys(cur).forEach((key) => {
        if (!acc[key]) {
          acc[key] = 0;
        }
        acc[key] += cur[key];
      });
      return acc;
    }, {});
  }

  sumPriceObject(taxes: any): number {
    let total = 0;
    for (const tax in taxes) {
      total += taxes[tax] ?? 0;
    }

    return total;
  }

  containsNullValue(obj) {
    if (typeof obj !== 'object' || obj === null) {
      return obj === null; // Return true if the value itself is null
    }

    for (const key in obj) {
      if (obj.hasOwnProperty(key)) {
        if (this.containsNullValue(obj[key])) {
          return true; // If any nested value is null, return true
        }
      }
    }

    return false; // If no null values are found, return false
  }

  withNullValue(mainObj, called = []) {
    const obj = { ...mainObj };
    for (const variable of called) {
      delete obj[variable];
    }
    if (typeof obj !== 'object' || obj === null) {
      return null; // Return null for non-object values
    }

    for (const key in obj) {
      if (obj.hasOwnProperty(key)) {
        if (obj[key] === null) {
          return key; // Found a null value, return the key
        } else if (typeof obj[key] === 'object') {
          const nestedKey = this.withNullValue(obj[key]);
          if (nestedKey !== null) {
            return `${key}.${nestedKey}`; // Found null in nested object, return nested key path
          }
        }
      }
    }

    return null; // No null values found, return null
  }
}
